<?php

// +----------------------------------------------------------------------+
// | Wollabot - A modular service bot for IRC channels written in PHP     |
// +----------------------------------------------------------------------+
// | Copyright (C) 2002 Christian Joergensen (mail@phpguru.dk)            |
// +----------------------------------------------------------------------+
// | This program is free software; you can redistribute it and/or modify |
// | it under the terms of the GNU General Public License as published by |
// | the Free Software Foundation; either version 2 of the License, or    |
// | (at your option) any later version.                                  |
// |                                                                      |
// | This program is distributed in the hope that it will be useful but   |
// | WITHOUT ANY WARRANTY; without even the implied warranty of           |
// | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the         |
// | GNU General Public License for more details                          |
// |                                                                      |
// | You should have received a copy of the GNU General Public License    |
// | along with this program; if not, write to the                        |
// |Free Software Foundation, Inc., 59 Temple Place                      |
// |Suite 330, Boston, MA  02111-1307, USA                               |
// +----------------------------------------------------------------------+
// | Author(s): Christian Jrgensen <mail@phpguru.dk>                     |
// +----------------------------------------------------------------------+
// 
// $Id: wollabot.php,v 1.5 2003/09/28 08:08:43 chrj Exp $                                                                 

define("WOLLABOT_E_WARNING", 1);
define("WOLLABOT_E_ERROR",   2);
define("WOLLABOT_E_FATAL",   4);
define("WOLLABOT_E_ALL",     7);

// Log types

define("WOLLABOT_LOG_FILE",      1);
define("WOLLABOT_LOG_STDOUT",    2);

// Misc.

define("WOLLABOT_DEBUG", false);

// Message types

define("WOLLABOT_TYPE_UNKNOWN",        1);
define("WOLLABOT_TYPE_CHANNEL",        2);
define("WOLLABOT_TYPE_QUERY",          4);
define("WOLLABOT_TYPE_CTCP",           8);
define("WOLLABOT_TYPE_NOTICE",         16);
define("WOLLABOT_TYPE_WHO",            32);
define("WOLLABOT_TYPE_JOIN",           64);
define("WOLLABOT_TYPE_INVITE",         128);
define("WOLLABOT_TYPE_ACTION",         256);
define("WOLLABOT_TYPE_TOPICCHANGE",    512);
define("WOLLABOT_TYPE_NICKCHANGE",     1024);
define("WOLLABOT_TYPE_KICK",           2048);
define("WOLLABOT_TYPE_QUIT",           4096);
define("WOLLABOT_TYPE_LOGIN",          8192);
define("WOLLABOT_TYPE_INFO",           16384);
define("WOLLABOT_TYPE_LIST",           32768);
define("WOLLABOT_TYPE_NAME",           65536);
define("WOLLABOT_TYPE_MOTD",           131072);
define("WOLLABOT_TYPE_MODECHANGE",     262144);
define("WOLLABOT_TYPE_PART",           524288);
define("WOLLABOT_TYPE_ERROR",          1048576);
define("WOLLABOT_TYPE_BANLIST",        2097152);
define("WOLLABOT_TYPE_TOPIC",          4194304);
define("WOLLABOT_TYPE_NONRELEVANT",    8388608);
define("WOLLABOT_TYPE_WHOIS",          16777216);
define("WOLLABOT_TYPE_WHOWAS",         33554432);
define("WOLLABOT_TYPE_USERMODE",       67108864);
define("WOLLABOT_TYPE_CHANNELMODE",    134217728);
define("WOLLABOT_TYPE_CTCP_REQUEST",   268435456);
define("WOLLABOT_TYPE_CTCP_REPLY",     536870912);
define("WOLLABOT_TYPE_ALL",            1073741823);

// inf timelimit

set_time_limit(0);

// core object follows

define("WOLLABOT_VERSION", "0.3.0");
define("WOLLABOT_VERSION_STRING", "Wollabot ".WOLLABOT_VERSION." - http://wollabot.sf.net");

// using SmartIRC API
// this check should be done in a try, catch
@require_once "Net/SmartIRC.php";
if (!defined("SMARTIRC_DEBUG_ALL")) {
  die("Wollabot needs the SmartIRC libary to be installed - get it with 'pear install -a Net_SmartIRC'");
}
// some includes

require_once "lib/module.php";

class Wollabot {

  // vars used for settings
  var $errorlevel, 
  $logtype, 
  $logfile, 
  $irc, 
  $configuration, 
  $modules, 
  $registered_modules, 
  $current_channels;
  
  var $bound_prefixes, 
  $bound_chan, 
  $bound_onconnect, 
  $bound_types;

  function Wollabot () {

    // provide default values for internally used vars
    $this->errorlevel = WOLLABOT_E_ALL;
    $this->log_type   = WOLLABOT_LOG_STDOUT;
    $this->log_file   = "irc.log";

    // PHP version checking
    if ((!function_exists("version_compare") || (version_compare(phpversion(), "4.1.0") < 0))) {
      $this->raise_error("You need PHP 4.1.0+ in order to use Wollabot", WOLLABOT_E_FATAL);
    }

    // Create IRC object from SmartIRC
    $this->irc = &new Net_SmartIRC;

    // Set minor SmartIRC options
    $this->irc->setUseSockets(TRUE);
    $this->irc->setCtcpVersion(WOLLABOT_VERSION_STRING);
    $this->irc->setAutoRetry(FALSE);

    if (WOLLABOT_DEBUG) {
      $this->irc->setDebug(SMARTIRC_DEBUG_ALL);
    } else {
      $this->irc->setDebug(SMARTIRC_DEBUG_NONE);
    }

    // Register man parser
    $this->irc->registerActionHandler(SMARTIRC_TYPE_ALL, ".*", $this, "parse_incoming");

    // Read config file into $this->conf, $this->modules and $this->channels
    $this->read_config_file("wollabot.ini");

    $this->registered_modules = array();

  }

  function raise_error($msg, $type = WOLLABOT_E_WARNING) {

    // raise an error

    if ($this->errorlevel & $type) {
      
      switch($type) {

      case WOLLABOT_E_WARNING:
        $this->print_log("Warning: ".$msg);
        break;

      case WOLLABOT_E_ERROR:
        $this->print_log("Error: ".$msg);
        break;

      case WOLLABOT_E_FATAL:
        $this->print_log("Fatal: ".$msg);
        $this->print_log("Closing down...");
        $this->destroy();
        break;

      default:
        break;

      }

    } else {

      // ignore

    }

    return true;

  }

  function destroy () {
    if ($this->irc->_socket != NULL) $this->disconnect();
    $this->print_log("Wollabot is closing.");
    exit;
  }

	function print_log($msg) {
		// write data to given log destination
  	$msg = date("d/m/Y H:i:s") . " -!- " . $msg . "\n";
    switch($this->log_type) {
    	case WOLLABOT_LOG_FILE:
         // check if the log resource is open
        	if (!is_resource($this->log_resource)) {
        		$this->log_resource = fopen($this->log_file, "w");
        	}
        	fputs($this->log_resource, $msg);
        	return true;
    	break;
    
    	case WOLLABOT_LOG_STDOUT:
        // print dose print to stdout, so why use php://stdout
        	print $msg;
        	return true;
    		break;
    
    	default:
        	$this->raise_error("Unknown log type", WOLLABOT_E_FATAL);
        	return false;
    		break;
    }
	}
  

  function read_config_file($cfg) {

    if (!file_exists($cfg)) {
      $this->raise_error("Could not open configuration file '$cfg'", WOLLABOT_E_FATAL);
    }

    // reset line counter
    $i = 0;

    // read file
    $file_array = file($cfg);

    foreach ($file_array as $line) {
    
      // increase line counter
      $i++;
    
      if ($line[0] == ";") {
	
	// ignore comments
	
      } elseif (preg_match("/^\[([^]]+)\].*$/", trim($line), $regs)) {
	
	// matches a[category] from config file

	$category = strtolower($regs[1]);
	
      } elseif (preg_match("/^([^[:space:]]+)[^=]*=[^\"]*\"([^\"]*)\".*$/", trim($line), $regs)) {
	
	// matches as setting: flaf = "foo"

	switch($category) {
	  
	  // put the configurations into the verious places where it belongs

	case "main":
	  $this->configuration[$regs[1]] = $regs[2];
	  break;
	  
	case "modules":
	  if (strtolower($regs[1]) == "loadmodule") {
	    
	    if (!file_exists("modules/".$regs[2])) {
	      $this->print_log("Could not open module '".$regs[2]."' - check your modules/ directory.", WOLLABOT_E_ERROR);
	    } else {
	      if (!is_array($this->modules)) $this->modules = array();
	      array_push($this->modules, $regs[2]);
	    }

	  } else {

	    $this->raise_error("Configuration error in 'module' section near line $i", WOLLABOT_E_FATAL);

	  }

	  break;
	  
	case "channel":

	  if (!isset($channel_name) && (strtolower($regs[1]) == "name")) {

	    // register new channel
	    $channel_name = $regs[2];

	  } elseif (isset($channel_name) && ((strtolower($regs[1]) == "on_join") || (strtolower($regs[1]) == "key"))) {

	    // set channel property (on_join|key)
	    $this->current_channels[$channel_name][$regs[1]] = $regs[2];

	    if (sizeOf($this->current_channels[$channel_name]) == 2) {
	      
	      // if both on_join and key is set unset channel name and wait for new channel
	      unset($channel_name);
	    }

	  } else {

	    $this->raise_error("Configuration error in 'channel' section near line $i", WOLLABOT_E_FATAL);
    
	  }

	  break;
	  
	}
	
      }
      
    }
    
  }

  function connect_and_register() {
    
    // Connects and sets up the connection with nick, username and realname

    $this->print_log("Connection to IRC server '".$this->configuration['server'].":".$this->configuration['port']."' ".
		     "with nickname '".$this->configuration['nickname']."' and username '".$this->configuration['username']."'");

    $this->irc->connect($this->configuration['server'], $this->configuration['port']);
    if (isset($this->configuration['passwd'])) {
    	$this->irc->login($this->configuration['nickname'], WOLLABOT_VERSION_STRING, 0, $this->configuration['username'], $this->configuration['passwd']);
    } else {
 		$this->irc->login($this->configuration['nickname'], WOLLABOT_VERSION_STRING, 0, $this->configuration['username']);
    }
  }

  function loop() {

    // Main loop - waiting for interactions

    $this->print_log("Entering main loop.");

    $this->irc->listen();

  }

  function disconnect() {

    // disconnects from IRC

    $this->irc->disconnect();

  }

  function data_to_parms ($data) {

    // reformats a 'data'-object form SmartIRC to Wollabot 'data'-array

    return array('from'     => $data->from,
		 'nick'             => $data->nick,
		 'ident'            => $data->ident,
		 'host'             => $data->host,
		 'channel'          => $data->channel,
		 'message'          => $data->message,
		 'message_exploded' => $data->messageex,
		 'type'             => $data->type,
		 'raw'              => $data->rawmessage,
		 'raw_exploded'     => $data->rawmessageex
		 );
  }

  function register_modules() {

    // Run through the modules from configuration and register them.
    // This is actually done using $wollabot->register_module(<classname>)
    // in the various modules - this method just triggers these calls
    // by including the files.

    $this->print_log("Registering modules...");
    if (is_array($this->modules)) foreach ($this->modules as $module) {

      $this->print_log("Reading '$module'...");

      // Make $wollabot a reference to the current Wollabot object
      $wollabot = &$this;

      include_once "modules/$module";

    }

  }

  function register_module($module_name) {

    // Instansiates the modules in the registered_modules array.
    // This is called from the modules themselves, triggerd when
    // included by register_modules()

    $this->print_log("Registering '$module_name'");
    $this->registered_modules[$module_name] = &new $module_name;

  }

  function prepare_modules() {

    // Prepares the modules by retrieving the various action-handlers
    // and storing them in central place.

    $this->print_log("Preparing modules");

    if (!is_array($this->registered_modules)) return false;

    foreach ($this->registered_modules as $module_name => $not_used) {
      
      $this->print_log("Preparing '$module_name'");

      // Set wollabot object
      $this->registered_modules[$module_name]->set_wollabot_object($this);
      
      // Set bound prefixes 
      if (is_array($this->registered_modules[$module_name]->bound_prefixes)) {
	foreach($this->registered_modules[$module_name]->bound_prefixes as $prefix => $methods) {
	  foreach($methods as $method) {

	    $usermethod = array('modulename' => $module_name, 'method' => $method);
	    
	    if (is_array($this->bound_prefixes[$prefix])) array_push($this->bound_prefixes[$prefix], $usermethod);
	    else $this->bound_prefixes[$prefix] = array($usermethod);

	    $this->print_log("Preparing prefix event: $module_name -> $method ");
   	    
	  } // foreach
	} // foreach
      } // if
      
      // Set bound channels 
      if (is_array($this->registered_modules[$module_name]->bound_chan)) {
	foreach($this->registered_modules[$module_name]->bound_chan as $channel => $methods) {
	  foreach($methods as $method) {
	    
	    $usermethod = array('modulename' => $module_name, 'method' => $method);
	    
	    if (is_array($this->bound_chan[$channel])) array_push($this->bound_chan[$channel], $usermethod);
	    else $this->bound_chan[$channel] = array($usermethod);

	    $this->print_log("Preparing channel event: $module_name -> $method ");
   	    
	  } // foreach
	} // foreach
      } // if
      
      // Set bound onconnect
      if (is_array($this->registered_modules[$module_name]->bound_onconnect)) {
	foreach($this->registered_modules[$module_name]->bound_onconnect as $notused => $method) {
	  
	  $usermethod = array('modulename' => $module_name, 'method' => $method);
	  
	  if (is_array($this->bound_onconnect)) array_push($this->bound_onconnect, $usermethod);
	  else $this->bound_onconnect = array($usermethod);
	  
	  $this->print_log("Preparing onconnect event: $module_name -> $method ");
    
	} // foreach
      } // if

      // Set bound types
      if (is_array($this->registered_modules[$module_name]->bound_types)) {
	foreach($this->registered_modules[$module_name]->bound_types as $type => $methods) {
	  foreach($methods as $method) {
	    
	    $usermethod = array('modulename' => $module_name, 'method' => $method);
	    
	    if (is_array($this->bound_types[$types])) array_push($this->bound_types[$type], $usermethod);
	    else $this->bound_types[$type] = array($usermethod);

	    $this->print_log("Preparing type event: $module_name -> $method ");
   	    
	  } // foreach
	} // foreach
      } // if


      // Set bound timers
      if (is_array($this->registered_modules[$module_name]->bound_frequencies)) {
	foreach($this->registered_modules[$module_name]->bound_frequencies as $frequency => $methods) {
	  foreach ($methods as $method) {

	    $this->irc->registerTimehandler($frequency, $this->registered_modules[$module_name], $method);
	    $this->print_log("Preparing timer event: $module_name -> $method ");
	  
	  } // foreach
	} // foreach
      } // if


    } // foreach
  } // function

  function parse_incoming(&$irc, &$data) {

    // Main function for parsing *ALL* incoming data. Triggers
    // various functions named check_* when an action is triggered
    // and these functions looks for bound methods to these triggers

    $data = $this->data_to_parms($data);

    // Check for prefixes is type is channel, query or notice
    if ($data['type'] & (WOLLABOT_TYPE_CHANNEL | WOLLABOT_TYPE_QUERY |WOLLABOT_TYPE_NOTICE)) 
      $this->check_prefixes($data);

    // Run onconnect if end of motd
    if ($data['raw_exploded'][1] == '376') 
      $this->check_onconnect($data);

    // Check channels
    if ($data['type'] == WOLLABOT_TYPE_CHANNEL) 
      $this->check_channel($data);
      
    if (is_array($this->bound_types)) {
    	foreach ($this->bound_types as $type => $methods) {
	  		if ($type & $data['type']) {
	    		foreach ($methods as $method) 
	    			$this->call_module_method($method['modulename'], $method['method'], array($data));
	  		}
      	}
    }
  }
  
  function check_onconnect($data) {
  
    // Checks for onconnect triggers

    $this->print_log("Checking for on-connect events.");

    if (!is_array($this->bound_onconnect)) return false;
    
    foreach ($this->bound_onconnect as $method) {
      $this->call_module_method($method['modulename'], $method['method'], array($data));
    }
  }


  function check_prefixes($data) {

    // Checks for prefix triggers

    if (is_array($this->bound_prefixes)) {
      foreach($this->bound_prefixes as $prefix => $methods) {
		if (substr($data['message'], 0, strlen($prefix)) == $prefix) {
	  		foreach($methods as $method => $info) 
	    		$this->call_module_method($info['modulename'], $info['method'], array($data));
			}
      	}
    }
  }



  function check_channel($data) {

    // Check channel trigger

  	if (is_array($this->bound_chan)) {
    	foreach($this->bound_chan as $channel => $methods) {
			if ($data['channel'] == $channel || $channel == 'all') {
	  			foreach($methods as $method => $info) 
	    			$this->call_module_method($info['modulename'], $info['method'], array($data));
			}
      	}
    }
  }



  function call_module_method($modulename, $method, $parms) {
    // Call module method with data array
    $this->print_log("Calling: $modulename -> $method");
    call_user_func_array(array(&$this->registered_modules[$modulename], $method), $parms);
  }
}


$wollabot = &new Wollabot;

?>